/* $Id: heart.c 2 2008-09-21 01:55:30Z gophi $ */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <ppm.h>

static void fasthelp(void)
{
	fprintf(stderr, "heart - generate heart image\n");
	fprintf(stderr, "This program shows how to use libppm library\n");
	fprintf(stderr, "Syntax: heart <outfile.ppm>|- <width> <height> <hxstep> <fstep> <angle>\n");
	fprintf(stderr, "Example: heart heart.ppm 640 480 0.001 0.05\n");
}

static int check_version(void)
{
	if (ppm_version_major() != PPM_VERSION_MAJOR)
	{
		fprintf(stderr, "Library version mismatch. Please recompile.\n");
		return -1;
	}

	fprintf(stderr, "Compiled with libppm v%u.%u\n", PPM_VERSION_MAJOR, PPM_VERSION_MINOR);
	fprintf(stderr, "Using libppm v%u.%u\n", ppm_version_major(), ppm_version_minor());

	return 0;
}

static ppm_err_t save_ppm(const ppm_t *ppm, const char *name)
{
	ppm_err_t err;

	if (!strcmp(name, "-"))
	{
		fprintf(stderr, "Writing ppm to stdout\n");
		err = ppm_save_fp(ppm, stdout);
	}
	else
	{
		fprintf(stderr, "Writing ppm to file %s\n", name);
		err = ppm_save_file(ppm, name);
	}

	return err;
}

static ppm_err_t draw_heart(ppm_t *ppm, double hxstep, double fstep)
{
	double hx;

	for (hx = -1; hx < 1; hx += hxstep)
	{
		double f;

		for (f = 0; f < 2 * sqrt(1 - hx * hx); f += fstep)
		{
			ppm_pixel_t *pixel;
			ppm_err_t err;
			int c = 0;

			uint16_t x = (hx + 1) * ppm->width / 2;
			uint16_t y = ((cbrt(hx * hx) + f - sqrt(1 - hx * hx) + 1) * (ppm->height / 2.5));

			if (y >= ppm->height)
				continue;

			y = ppm->height - y - 1;

			if ((err = ppm_pixel(ppm, &pixel, x, y)) != PPM_ERR_OK)
				return err;

			pixel->b.rgb[0] = 255;
			pixel->b.rgb[1] = c;
			pixel->b.rgb[2] = c;
		}
	}

	return PPM_ERR_OK;
}

int main(int ac, char * const av[])
{
	const char *outf;
	uint16_t width, height;
	double hxstep, fstep;
	ppm_t ppm;
	ppm_err_t err;

	if (ac != 6)
	{
		fasthelp();
		return EXIT_FAILURE;
	}

	outf = av[1];
	width = atoi(av[2]);
	height = atoi(av[3]);
	hxstep = atof(av[4]);
	fstep = atof(av[5]);

	if (check_version() == -1)
		return EXIT_FAILURE;

	if ((err = ppm_create(&ppm, width, height, PPM_BPP_24)) != PPM_ERR_OK)
		goto fail;

	if ((err = draw_heart(&ppm, hxstep, fstep)) != PPM_ERR_OK)
		goto fail;

	if ((err = save_ppm(&ppm, outf)) != PPM_ERR_OK)
		goto fail;

	if ((err = ppm_destroy(&ppm)) != PPM_ERR_OK)
		goto fail;

	return EXIT_SUCCESS;

fail:
	ppm_perror(err, av[0]);
	return EXIT_FAILURE;
}
