/* $Id: histogram.c 2 2008-09-21 01:55:30Z gophi $ */

#include <inttypes.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ppm.h>

static const uint16_t HIST_HEIGHT = 192;

static ppm_err_t load_ppm(ppm_t *ppm, const char *name)
{
	ppm_err_t err;

	if (!strcmp(name, "-"))
	{
		fprintf(stderr, "Reading ppm from stdin\n");
		err = ppm_load_fp(ppm, stdin);
	}
	else
	{
		fprintf(stderr, "Reading ppm from file %s\n", name);
		err = ppm_load_file(ppm, name);
	}

	if (err != PPM_ERR_OK)
		return err;

	assert(ppm->magic == PPM_MAGIC && (ppm->bpp == PPM_BPP_24 || ppm->bpp == PPM_BPP_48));
	fprintf(stderr, "Read %u-bit ppm %ux%u\n", (ppm->bpp == PPM_BPP_24) ? 24 : 48, ppm->width, ppm->height);

	return err;
}

static ppm_err_t save_ppm(const ppm_t *ppm, const char *name)
{
	ppm_err_t err;

	if (!strcmp(name, "-"))
	{
		fprintf(stderr, "Writing ppm to stdout\n");
		err = ppm_save_fp(ppm, stdout);
	}
	else
	{
		fprintf(stderr, "Writing ppm to file %s\n", name);
		err = ppm_save_file(ppm, name);
	}

	return err;
}

static void show_fasthelp(void)
{
	fprintf(stderr, "histogram - create histogram for a ppm image\n");
	fprintf(stderr, "This program shows how to use libppm library\n");
	fprintf(stderr, "Syntax: histogram <input.ppm>|- <output.ppm>|-\n");
}

static int check_version(void)
{
	if (ppm_version_major() != PPM_VERSION_MAJOR)
	{
		fprintf(stderr, "Library version mismatch. Please recompile.\n");
		return -1;
	}

	fprintf(stderr, "Compiled with libppm v%u.%u\n", PPM_VERSION_MAJOR, PPM_VERSION_MINOR);
	fprintf(stderr, "Using libppm v%u.%u\n", ppm_version_major(), ppm_version_minor());

	return 0;
}

static ppm_err_t draw_hline(ppm_t *ppm, uint16_t x, uint16_t y0, uint16_t y1, const ppm_pixel_t *color)
{
	uint16_t y;

	for (y = y0; y <= y1; ++y)
	{
		ppm_pixel_t *pixel;
		ppm_err_t err;

		if ((err = ppm_pixel(ppm, &pixel, x, y)) != PPM_ERR_OK)
			return err;

		memcpy(pixel, color, sizeof(*pixel));
	}

	return PPM_ERR_OK;
}

static ppm_err_t draw_vline(ppm_t *ppm, uint16_t x0, uint16_t x1, uint16_t y, const ppm_pixel_t *color)
{
	uint16_t x;

	for (x = x0; x <= x1; ++x)
	{
		ppm_pixel_t *pixel;
		ppm_err_t err;

		if ((err = ppm_pixel(ppm, &pixel, x, y)) != PPM_ERR_OK)
			return err;

		memcpy(pixel, color, sizeof(*pixel));
	}

	return PPM_ERR_OK;
}

static ppm_err_t count_subpixels(const ppm_t *in, size_t index, size_t table[256])
{
	size_t x, y;

	memset(table, 0, sizeof(table[0]) * 256);
	for (y = 0; y < in->height; ++y)
	{
		for (x = 0; x < in->width; ++x)
		{
			const ppm_pixel_t *pixel;
			ppm_pixel_t pixel24;
			ppm_err_t err;

			if ((err = ppm_pixel_const(in, &pixel, x, y)) != PPM_ERR_OK)
				return err;

			if (in->bpp == PPM_BPP_48)
			{
				pixel24.b.rgb[0] = pixel->w.rgb[0] / 256;
				pixel24.b.rgb[1] = pixel->w.rgb[1] / 256;
				pixel24.b.rgb[2] = pixel->w.rgb[2] / 256;

				pixel = &pixel24;
			}

			++table[pixel->b.rgb[index]];
		}
	}

	return PPM_ERR_OK;
}

static size_t calculate_max(size_t table[3][256])
{
	size_t max = 0;
	size_t i, j;

	for (i = 0; i < 3; ++i)
		for (j = 0; j < 256; ++j)
			if (table[i][j] > max)
				max = table[i][j];

	fprintf(stderr, "Maximum count %u\n", max);
	return max;
}

static ppm_err_t draw_value(ppm_t *out, uint16_t x, const ppm_pixel_t *color, size_t value, size_t max)
{
	uint16_t y;
	for (y = out->height - (value * HIST_HEIGHT / max) - 1; y < out->height; ++y)
	{
		ppm_pixel_t *pixel;
		ppm_err_t err;

		if ((err = ppm_pixel(out, &pixel, x, y)) != PPM_ERR_OK)
			return err;

		pixel->b.rgb[0] |= color->b.rgb[0];
		pixel->b.rgb[1] |= color->b.rgb[1];
		pixel->b.rgb[2] |= color->b.rgb[2];
	}

	return PPM_ERR_OK;
}

static ppm_err_t draw_channel(ppm_t *out, size_t index, size_t table[256], size_t max)
{
	ppm_pixel_t color;
	uint16_t x;

	memset(&color, 0, sizeof(color));
	color.b.rgb[index] = 0xFF;

	for (x = 0; x < 256; ++x)
	{
		ppm_err_t err;
		if ((err = draw_value(out, x, &color, table[x], max)) != PPM_ERR_OK)
			return err;
	}

	return PPM_ERR_OK;
}

static ppm_err_t create_histogram(ppm_t *out, const ppm_t *in)
{
	static const ppm_pixel_t grey = { .b.rgb = { 0x20, 0x20, 0x20 } };
	size_t i;
	ppm_err_t err;
	size_t values[3][256];
	size_t max;

	for (i = 0; i < out->width; i += 8)
		if ((err = draw_hline(out, i, 0, out->height - 1, &grey)) != PPM_ERR_OK)
			return err;

	for (i = 0; i < out->height; i += 8)
		if ((err = draw_vline(out, 0, out->width - 1, i, &grey)) != PPM_ERR_OK)
			return err;

	for (i = 0; i < 3; ++i)
		if ((err = count_subpixels(in, i, values[i])) != PPM_ERR_OK)
			return err;

	max = calculate_max(values);
	for (i = 0; i < 3; ++i)
		if ((err = draw_channel(out, i, values[i], max)) != PPM_ERR_OK)
			return err;

	return PPM_ERR_OK;
}

int main(int ac, char * const av[])
{
	const char *inf, *outf;
	ppm_t inppm, outppm;
	ppm_err_t err;

	if (ac != 3)
	{
		show_fasthelp();
		return EXIT_FAILURE;
	}

	inf = av[1];
	outf = av[2];

	if (check_version() == -1)
		return EXIT_FAILURE;

	if ((err = load_ppm(&inppm, inf)) != PPM_ERR_OK)
		goto fail;

	if ((err = ppm_create(&outppm, 256, HIST_HEIGHT, PPM_BPP_24)) != PPM_ERR_OK)
		goto fail;

	if ((err = create_histogram(&outppm, &inppm)) != PPM_ERR_OK)
		goto fail;

	if ((err = save_ppm(&outppm, outf)) != PPM_ERR_OK)
		goto fail;

	if ((err = ppm_destroy(&outppm)) != PPM_ERR_OK)
		goto fail;

	if ((err = ppm_destroy(&inppm)) != PPM_ERR_OK)
		goto fail;

	return EXIT_SUCCESS;

fail:
	ppm_perror(err, av[0]);
	return EXIT_FAILURE;
}
