/* $Id: normalize.c 2 2008-09-21 01:55:30Z gophi $ */

#include <inttypes.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ppm.h>

static void setmax24(uint16_t *max, const ppm_pixel_t *pixel)
{
	size_t i;
	for (i = 0; i < 3; ++i)
		if (pixel->b.rgb[i] > *max)
			*max = pixel->b.rgb[i];
}

static void setmax48(uint16_t *max, const ppm_pixel_t *pixel)
{
	size_t i;
	for (i = 0; i < 3; ++i)
		if (pixel->w.rgb[i] > *max)
			*max = pixel->w.rgb[i];
}

static void apply24(ppm_pixel_t *pixel, double factor)
{
	pixel->b.rgb[0] *= factor;
	pixel->b.rgb[1] *= factor;
	pixel->b.rgb[2] *= factor;
}

static void apply48(ppm_pixel_t *pixel, double factor)
{
	pixel->w.rgb[0] *= factor;
	pixel->w.rgb[1] *= factor;
	pixel->w.rgb[2] *= factor;
}

static ppm_err_t load_ppm(ppm_t *ppm, const char *name)
{
	ppm_err_t err;

	if (!strcmp(name, "-"))
	{
		fprintf(stderr, "Reading ppm from stdin\n");
		err = ppm_load_fp(ppm, stdin);
	}
	else
	{
		fprintf(stderr, "Reading ppm from file %s\n", name);
		err = ppm_load_file(ppm, name);
	}

	if (err != PPM_ERR_OK)
		return err;

	assert(ppm->magic == PPM_MAGIC && (ppm->bpp == PPM_BPP_24 || ppm->bpp == PPM_BPP_48));
	fprintf(stderr, "Read %u-bit ppm %ux%u\n", (ppm->bpp == PPM_BPP_24) ? 24 : 48, ppm->width, ppm->height);

	return err;
}

static ppm_err_t save_ppm(const ppm_t *ppm, const char *name)
{
	ppm_err_t err;

	if (!strcmp(name, "-"))
	{
		fprintf(stderr, "Writing ppm to stdout\n");
		err = ppm_save_fp(ppm, stdout);
	}
	else
	{
		fprintf(stderr, "Writing ppm to file %s\n", name);
		err = ppm_save_file(ppm, name);
	}

	return err;
}

static ppm_err_t calculate_factor(const ppm_t *ppm, double *factor)
{
	void (*setmax)(uint16_t *max, const ppm_pixel_t *pixel) = (ppm->bpp == PPM_BPP_24) ? setmax24 : setmax48;
	uint16_t max = 0;
	uint16_t x, y;

	fprintf(stderr, "Calculating normalize factor using slow ppm_pixel mode\n");
	for (y = 0; y < ppm->height; ++y)
	{
		for (x = 0; x < ppm->width; ++x)
		{
			const ppm_pixel_t *pixel;
			ppm_err_t err;

			if ((err = ppm_pixel_const(ppm, &pixel, x, y)) != PPM_ERR_OK)
				return err;

			setmax(&max, pixel);
		}
	}

	if (max == 0)
		fprintf(stderr, "Warning: Normalizing black image\n");

	*factor = max ? ((ppm->bpp == PPM_BPP_24) ? 255. : 65535.) / max : 1;
	return PPM_ERR_OK;
}

static void apply_factor(ppm_t *ppm, double factor)
{
	void (*apply)(ppm_pixel_t *pixel, double factor) = (ppm->bpp == PPM_BPP_24) ? apply24 : apply48;
	size_t i;

	fprintf(stderr, "Applying factor %.2f to image using fast linear mode\n", factor);
	for (i = 0; i < (size_t) ppm->width * ppm->height; ++i)
		apply(&ppm->data[i], factor);
}

static void show_fasthelp(void)
{
	fprintf(stderr, "normalize - transform image to span full color range\n");
	fprintf(stderr, "This program shows how to use libppm library\n");
	fprintf(stderr, "Syntax: normalize <input.ppm>|- <output.ppm>|-\n");
}

static int check_version(void)
{
	if (ppm_version_major() != PPM_VERSION_MAJOR)
	{
		fprintf(stderr, "Library version mismatch. Please recompile.\n");
		return -1;
	}

	fprintf(stderr, "Compiled with libppm v%u.%u\n", PPM_VERSION_MAJOR, PPM_VERSION_MINOR);
	fprintf(stderr, "Using libppm v%u.%u\n", ppm_version_major(), ppm_version_minor());

	return 0;
}

int main(int ac, char * const av[])
{
	const char *inf, *outf;
	ppm_t ppm;
	ppm_err_t err;
	double factor;

	if (ac != 3)
	{
		show_fasthelp();
		return EXIT_FAILURE;
	}

	inf = av[1];
	outf = av[2];

	if (check_version() == -1)
		return EXIT_FAILURE;

	if ((err = load_ppm(&ppm, inf)) != PPM_ERR_OK)
		goto fail;

	if ((err = calculate_factor(&ppm, &factor)) != PPM_ERR_OK)
		goto fail;

	apply_factor(&ppm, factor);

	if ((err = save_ppm(&ppm, outf)) != PPM_ERR_OK)
		goto fail;

	if ((err = ppm_destroy(&ppm)) != PPM_ERR_OK)
		goto fail;

	return EXIT_SUCCESS;

fail:
	ppm_perror(err, av[0]);
	return EXIT_FAILURE;
}
